import datetime

from flask_restplus import Namespace, fields, Resource, reqparse

from werkzeug.datastructures import FileStorage

from . import api, DatetimeToTimestamp
from drucker_dashboard import DruckerDashboardClient
from drucker_dashboard.models import db, Service, Model


mdl_info_namespace = Namespace('models', description='Model Endpoint.')
success_or_not = mdl_info_namespace.model('Success', {
    'status': fields.Boolean(
        required=True
    ),
    'message': fields.String(
        required=True
    )
})
mdl_info = mdl_info_namespace.model('Model', {
    'model_id': fields.Integer(
        readOnly=True,
        description='Model ID.'
    ),
    'application_id': fields.Integer(
        readOnly=True,
        description='Application ID.'
    ),
    'model_path': fields.String(
        readOnly=True,
        description='Model file path.',
        example='/ml-1234567.model'
    ),
    'register_date': DatetimeToTimestamp(
        readOnly=True,
        description='Register date.'
    ),
    'description': fields.String(
        required=False,
        description='Description.',
        example='This is a sample.'
    )
})

@mdl_info_namespace.route('/<int:application_id>/models')
class ApiApplicationIdModels(Resource):
    upload_model_parser = reqparse.RequestParser()
    upload_model_parser.add_argument('file', location='files',
                                     type=FileStorage, required=True)
    upload_model_parser.add_argument('description', type=str, required=False, location='form')

    @mdl_info_namespace.marshal_list_with(mdl_info)
    def get(self, application_id:int):
        """get_models"""
        return Model.query.filter_by(application_id=application_id).all()

    @mdl_info_namespace.marshal_with(success_or_not)
    @mdl_info_namespace.expect(upload_model_parser)
    def post(self, application_id:int):
        """upload_model"""
        args = self.upload_model_parser.parse_args()
        file = args['file']
        description = args['description']
        model_path = "ml-{0:%Y%m%d%H%M%S}.model".format(datetime.datetime.utcnow())

        sobj = db.session.query(Service).filter(Service.application_id == application_id).first()
        host = sobj.host
        drucker_dashboard_application = DruckerDashboardClient(logger=api.logger, host=host)
        response_body = drucker_dashboard_application.run_upload_model(model_path, file)
        if not response_body.get("status", True):
            raise Exception(response_body.get("message", "Error."))

        mobj = Model(application_id=application_id,
                     model_path=model_path,
                     description=description)
        db.session.add(mobj)
        db.session.commit()
        db.session.close()
        return response_body

@mdl_info_namespace.route('/<int:application_id>/models/<int:model_id>')
class ApiApplicationIdModelId(Resource):
    update_config_parser = reqparse.RequestParser()
    update_config_parser.add_argument('description', type=str, required=True, location='form')

    @mdl_info_namespace.marshal_with(mdl_info)
    def get(self, application_id:int, model_id:int):
        """get_model"""
        return Model.query.filter_by(
            application_id=application_id,
            model_id=model_id).first_or_404()

    @mdl_info_namespace.marshal_with(success_or_not)
    @mdl_info_namespace.expect(update_config_parser)
    def patch(self, application_id:int, model_id:int):
        """update_model"""
        args = self.update_config_parser.parse_args()
        description = args['description']

        mobj = db.session.query(Model).filter(
            Model.application_id==application_id,
            Model.model_id==model_id).one()
        mobj.description = description
        response_body = {"status": True, "message": "Success."}
        db.session.commit()
        db.session.close()
        return response_body

    @mdl_info_namespace.marshal_with(success_or_not)
    def delete(self, application_id:int, model_id:int):
        """delete_model"""
        mobj = db.session.query(Model).filter(
            Model.application_id==application_id,
            Model.model_id==model_id).one_or_none()
        if mobj is None:
            raise Exception("No such model_id.")
        db.session.query(Model).filter(
            Model.application_id==application_id,
            Model.model_id==model_id).delete()
        response_body = {"status": True, "message": "Success."}
        db.session.commit()
        db.session.close()
        return response_body



import datetime
import json

from flask_restplus import Namespace, fields, Resource, reqparse

from . import api, DatetimeToTimestamp
from .api_kubernetes import update_dbs_kubernetes, switch_drucker_service_model_assignment
from drucker_dashboard import DruckerDashboardClient
from drucker_dashboard.models import db, Kubernetes, Application, Service, EvaluationResult, Evaluation


srv_info_namespace = Namespace('services', description='Service Endpoint.')
success_or_not = srv_info_namespace.model('Success', {
    'status': fields.Boolean(
        required=True
    ),
    'message': fields.String(
        required=True
    )
})
srv_info = srv_info_namespace.model('Service', {
    'service_id': fields.Integer(
        readOnly=True,
        description='Service ID.'
    ),
    'service_name': fields.String(
        required=True,
        description='Service tag.',
        example='drucker-sample-development-123456789'
    ),
    'display_name': fields.String(
        required=True,
        description='Display name.',
        example='dev-001'
    ),
    'application_id': fields.Integer(
        readOnly=True,
        description='Application ID.'
    ),
    'model_id': fields.Integer(
        readOnly=True,
        description='Model ID.'
    ),
    'service_level': fields.String(
        required=True,
        description='Service level. [development/beta/staging/sandbox/production]',
        example='development'
    ),
    'register_date': DatetimeToTimestamp(
        readOnly=True,
        description='Register date.'
    ),
    'confirm_date': DatetimeToTimestamp(
        readOnly=True,
        description='Existance confirmation date.'
    ),
    'update_date': DatetimeToTimestamp(
        readOnly=True,
        description='Update date.'
    ),
    'host': fields.String(
        required=False,
        description='host.',
        example='drucker-sample.example.com'
    ),
    'description': fields.String(
        required=False,
        description='Description.',
        example='This is a sample.'
    )
})


@srv_info_namespace.route('/<int:application_id>/services')
class ApiApplicationIdServices(Resource):
    @srv_info_namespace.marshal_list_with(srv_info)
    def get(self, application_id:int):
        """get_services"""
        return Service.query.filter_by(application_id=application_id).all()

@srv_info_namespace.route('/<int:application_id>/services/<int:service_id>')
class ApiApplicationIdServiceId(Resource):
    switch_model_parser = reqparse.RequestParser()
    switch_model_parser.add_argument('model_id', type=int, required=True, location='form')

    update_config_parser = reqparse.RequestParser()
    update_config_parser.add_argument('display_name', type=str, required=True, location='form')
    update_config_parser.add_argument('description', type=str, required=False, location='form')

    @srv_info_namespace.marshal_with(srv_info)
    def get(self, application_id:int, service_id:int):
        """get_service"""
        return Service.query.filter_by(
            application_id=application_id,
            service_id=service_id).first_or_404()

    @srv_info_namespace.marshal_with(success_or_not)
    @srv_info_namespace.expect(switch_model_parser)
    def put(self, application_id:int, service_id:int):
        """switch_service_model_assignment"""
        args = self.switch_model_parser.parse_args()
        model_id = args['model_id']
        response_body = switch_drucker_service_model_assignment(
            application_id, service_id, model_id)
        db.session.commit()
        db.session.close()
        return response_body

    @srv_info_namespace.marshal_with(success_or_not)
    @srv_info_namespace.expect(update_config_parser)
    def patch(self, application_id:int, service_id:int):
        """update_service_config"""
        args = self.update_config_parser.parse_args()
        display_name = args['display_name']
        description = args['description']

        sobj = db.session.query(Service).filter(
            Service.application_id == application_id,
            Service.service_id == service_id).one()
        sobj.display_name = display_name
        if description is not None:
            sobj.description = description
        sobj.update_date = datetime.datetime.utcnow()
        response_body = {"status": True, "message": "Success."}
        db.session.commit()
        db.session.close()
        return response_body

    @srv_info_namespace.marshal_with(success_or_not)
    def delete(self, application_id:int, service_id:int):
        """delete_service"""
        aobj = db.session.query(Application).filter(
            Application.application_id == application_id).one_or_none()
        if aobj is None:
            raise Exception("No such application_id.")
        sobj = db.session.query(Service).filter(
            Service.application_id == application_id,
            Service.service_id == service_id).one_or_none()
        if sobj is None:
            raise Exception("No such service_id.")

        if aobj.kubernetes_id is None:
            db.session.query(Service).filter(
                Service.application_id == application_id,
                Service.service_id == service_id).delete()
            db.session.flush()
        else:
            kubernetes_id = aobj.kubernetes_id
            kobj = db.session.query(Kubernetes).filter(
                Kubernetes.kubernetes_id == kubernetes_id).one_or_none()
            if kobj is None:
                raise Exception("No such kubernetes_id.")
            config_path = kobj.config_path
            from kubernetes import client, config
            config.load_kube_config(config_path)

            apps_v1 = client.AppsV1Api()
            v1_deployment = apps_v1.delete_namespaced_deployment(
                name="{0}-deployment".format(sobj.service_name),
                namespace=sobj.service_level,
                body=client.V1DeleteOptions()
            )
            core_vi = client.CoreV1Api()
            v1_service = core_vi.delete_namespaced_service(
                name="{0}-service".format(sobj.service_name),
                namespace=sobj.service_level,
                body=client.V1DeleteOptions()
            )
            extensions_v1_beta = client.ExtensionsV1beta1Api()
            v1_beta1_ingress = extensions_v1_beta.delete_namespaced_ingress(
                name="{0}-ingress".format(sobj.service_name),
                namespace=sobj.service_level,
                body=client.V1DeleteOptions()
            )
            autoscaling_v1 = client.AutoscalingV1Api()
            v1_horizontal_pod_autoscaler = autoscaling_v1.delete_namespaced_horizontal_pod_autoscaler(
                name="{0}-autoscaling".format(sobj.service_name),
                namespace=sobj.service_level,
                body=client.V1DeleteOptions()
            )
            applist = set((aobj.application_name,))
            update_dbs_kubernetes(kubernetes_id, applist)
        response_body = {"status": True, "message": "Success."}
        db.session.commit()
        db.session.close()
        return response_body

@srv_info_namespace.route('/<int:application_id>/services/<int:service_id>/evaluate')
class ApiEvaluate(Resource):
    eval_parser = reqparse.RequestParser()
    eval_parser.add_argument('evaluation_id', location='form', type=int, required=False)
    eval_parser.add_argument('overwrite', location='form', type=bool, required=False)

    @srv_info_namespace.expect(eval_parser)
    def post(self, application_id:int, service_id:int):
        """evaluate"""
        args = self.eval_parser.parse_args()
        eval_id = args.get('evaluation_id', None)
        if eval_id:
            eobj = Evaluation.query.filter_by(
                application_id=application_id,
                evaluation_id=eval_id).first_or_404()
        else:
            # if evaluation_id is not given, use the lastest one.
            eobj = Evaluation.query\
                .filter_by(application_id=application_id)\
                .order_by(Evaluation.register_date.desc()).first_or_404()

        sobj = Service.query.filter_by(
            application_id=application_id,
            service_id=service_id).first_or_404()

        robj = db.session.query(EvaluationResult)\
            .filter(EvaluationResult.model_id == sobj.model_id,
                    EvaluationResult.evaluation_id == eobj.evaluation_id).one_or_none()
        if robj is not None and args.get('overwrite', False):
            return json.loads(robj.result)

        eval_result_path = "eval-result-{0:%Y%m%d%H%M%S}.txt".format(datetime.datetime.utcnow())
        drucker_dashboard_application = DruckerDashboardClient(logger=api.logger, host=sobj.host)
        response_body = drucker_dashboard_application.run_evaluate_model(eobj.data_path, eval_result_path)

        if response_body['status']:
            result = json.dumps(response_body)
            if robj is None:
                robj = EvaluationResult(model_id=sobj.model_id,
                                        data_path=eval_result_path,
                                        evaluation_id=eobj.evaluation_id,
                                        result=result)
                db.session.add(robj)
            else:
                robj.data_path = eval_result_path
                robj.result = result
            db.session.commit()
            db.session.close()

        return response_body



import datetime
from .dao import db
from sqlalchemy import (
    Column, Integer, DateTime,
    String, Text, UniqueConstraint
)


class Model(db.Model):
    """
    Model Info
    """
    __tablename__ = 'models'
    __table_args__ = (
        UniqueConstraint('model_id'),
        UniqueConstraint('application_id','model_path'),
        {'mysql_engine': 'InnoDB'}
    )

    model_id = Column(Integer, primary_key=True, autoincrement=True)
    application_id = Column(Integer, nullable=False)
    model_path = Column(String(512), nullable=False)
    description = Column(Text, nullable=True)
    register_date = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)

    @property
    def serialize(self):
        return {
            'model_id': self.model_id,
            'application_id': self.application_id,
            'model_path': self.model_path,
            'description': self.description,
            'register_date': self.register_date.strftime('%Y-%m-%d %H:%M:%S'),
        }



import enum
from .dao import db
from sqlalchemy import (
    Column, Integer,
    Enum,
    ForeignKey,
)
from sqlalchemy.orm import relationship


class Role(enum.Enum):
    viewer = 1
    editor = 2
    owner = 3


class ApplicationUserRole(db.Model):
    """
    Application-User Role
    """
    __tablename__ = 'application_user_roles'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    application_id = Column(Integer, ForeignKey('applications.application_id'), primary_key=True)
    user_id = Column(Integer, ForeignKey('users.user_id'), primary_key=True)
    role = Column(Enum(Role), nullable=False, default=Role.viewer)

    application = relationship('Application', lazy='joined', innerjoin=True)
    user = relationship('User')



import ldap

from .authenticator import Authenticator


class LdapAuthenticator(Authenticator):
    def __init__(self, config, logger):
        self.host = config['host']
        self.port = config['port']
        self.bind_dn = config['bind_dn']
        self.bind_password = config['bind_password']
        self.search_filter = config['search_filter']
        self.search_base_dns = config['search_base_dns']
        self.logger = logger

    def auth_user(self, username, password):
        lobj = ldap.initialize('ldap://{}:{}'.format(self.host, self.port))
        try:
            lobj.bind_s(self.bind_dn, self.bind_password)
            filter_str = self.search_filter % username
            for search_baase_dn in self.search_base_dns:
                result = lobj.search_s(search_baase_dn, ldap.SCOPE_SUBTREE, filter_str)
                user_dn, user_attrs = result[0]
                lobj.simple_bind_s(user_dn, password)
                return {
                    'uid': username,
                    'name': user_attrs.get('givenName', [b''])[0].decode('utf-8')
                }
            return None
        except IndexError:
            self.logger.error('"{}" not found'.format(username))
            return None
        except ldap.LDAPError as e:
            self.logger.error(e)
            return None



# -*- coding: utf-8 -*-


import os
import types

from flask import Flask, render_template
from flask_cors import CORS

from drucker_dashboard.utils import DruckerDashboardConfig
from drucker_dashboard.models import db
from drucker_dashboard.apis import api
from drucker_dashboard.auth import auth


def create_app(config_file: str = "settings.yml", logger_file: str = None):
    """create_app."""

    app = Flask(__name__, static_folder='static', template_folder='static/dist')
    app.logger.disabled = True
    for h in app.logger.handlers[:]:
        app.logger.removeHandler(h)

    @app.route('/')
    @app.route('/<path:path>')
    def root_url(**kwargs):
        return render_template('index.html')

    # load configurations
    config = DruckerDashboardConfig(config_file)
    if not os.path.isdir(config.DIR_KUBE_CONFIG):
        os.makedirs(config.DIR_KUBE_CONFIG)

    # set configurations
    app.config['SQLALCHEMY_DATABASE_URI'] = config.DB_URL
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = True
    app.config['DEBUG'] = os.getenv("FLASK_DEBUG", "True")
    app.config['SWAGGER_UI_DOC_EXPANSION'] = 'list'
    app.config['RESTPLUS_VALIDATE'] = True
    app.config['MAX_CONTENT_LENGTH'] = int(os.getenv("FLASK_MAX_CONTENT_LENGTH", "1073741824"))

    # logger
    try:
        l = types.ModuleType('logger')
        with open(logger_file, mode='rb') as f:
            exec(compile(f.read(), logger_file, 'exec'), l.__dict__)
        logger = l.__dict__.get("logger")
        logger.info("Use custom logger.")
    except Exception as e:
        from drucker_dashboard.logger import logger
        logger.info(str(e))
        logger.info("Invalid logger.")
        logger.info("Use default logger.")

    # initialize applications
    api.init_app(app, dashboard_config=config, logger=logger)
    auth.init_app(app, api, config.AUTH_CONFIG, logger=logger)
    CORS(app)
    db.init_app(app)
    db.create_all(app=app)

    return app


def main() -> None:
    app = create_app()
    app.run(host='0.0.0.0', port=18080, threaded=True)


if __name__ == '__main__':
    main()



# coding: utf-8


import traceback

from functools import wraps
from flask import jsonify, request, abort
from flask_jwt_simple import JWTManager, create_jwt, get_jwt_identity, jwt_required
from jwt.exceptions import PyJWTError
from flask_jwt_simple.exceptions import InvalidHeaderError, NoAuthorizationError

from .ldap import LdapAuthenticator
from .exceptions import ApplicationUserRoleException
from .authenticator import EmptyAuthenticator
from drucker_dashboard.models import db, User, Application, ApplicationUserRole, Role


class Auth(object):
    def __init__(self):
        self.__enabled = False
        self.authenticator = EmptyAuthenticator()
        self.logger = None

    def is_enabled(self):
        return self.__enabled

    def init_app(self, app, api, auth_conf, logger):
        JWTManager(app)
        self.logger = logger

        if auth_conf is None:
            self.__enabled = False
            self.authenticator = EmptyAuthenticator()
        else:
            self.__enabled = True
            app.config['JWT_SECRET_KEY'] = auth_conf['secret']
            if 'ldap' in auth_conf:
                self.authenticator = LdapAuthenticator(auth_conf['ldap'], logger)

        # Add endpoints
        @app.route('/api/login', methods=['POST'])
        def login():
            params = request.get_json()
            username = params.get('username', None)
            password = params.get('password', None)
            user_info = self.authenticator.auth_user(username, password)
            if user_info is not None:
                user_id = self.user(user_info)
                ret = {'jwt': create_jwt(identity=user_id)}
                return jsonify(ret), 200
            else:
                return jsonify({'message': 'Authentication failed'}), 401

        @app.route('/api/credential', methods=['GET'])
        @jwt_required
        def credential():
            user_id = get_jwt_identity()
            uobj = db.session.query(User).filter(User.user_id == user_id).one_or_none()
            if uobj is None:
                abort(404)
            application_roles = uobj.applications
            applications = [{'application_id': ar.application_id, 'role': ar.role.name} for ar in application_roles]
            # applications which don't have users are also accesssible as owner
            application_ids = db.session.query(ApplicationUserRole.application_id).distinct().all()
            ids = [application_id for application_id, in application_ids]
            public_applications = db.session.query(Application).filter(~Application.application_id.in_(ids)).all()
            applications += [
                {'application_id': app.application_id, 'role': Role.owner.name} for app in public_applications]

            return jsonify({'user': uobj.serialize, 'applications': applications}), 200

        @api.errorhandler(NoAuthorizationError)
        @api.errorhandler(InvalidHeaderError)
        @api.errorhandler(PyJWTError)
        @api.errorhandler(ApplicationUserRoleException)
        def authorization_error_handler(error):
            self.logger.error(error)
            self.logger.error(traceback.format_exc())
            return {'message': 'Authorization failed'}, 401

    def user(self, user_info):
        uobj = db.session.query(User).filter(User.auth_id == user_info['uid']).one_or_none()
        if uobj is not None:
            return uobj.user_id

        uobj = User(auth_id=user_info['uid'],
                    user_name=user_info['name'])
        db.session.add(uobj)
        db.session.flush()
        db.session.commit()
        user_id = uobj.user_id
        db.session.close()
        return user_id


auth = Auth()


def fetch_role(user_id, application_id):
    role = db.session.query(ApplicationUserRole).filter(
        ApplicationUserRole.application_id == application_id,
        ApplicationUserRole.user_id == user_id).one_or_none()
    if role is None:
        # applications which don't have users are also accesssible as owner
        roles = db.session.query(ApplicationUserRole).filter(
            ApplicationUserRole.application_id == application_id).count()
        if roles == 0:
            return Role.owner
        else:
            return None
    else:
        return role.role


def auth_required(fn):
    def check_role(user_id, application_id, method):
        role = fetch_role(user_id, application_id)
        if role is None:
            return False
        if method == 'GET':
            return True
        elif role == Role.editor or role == Role.owner:
                return True
        return False

    @wraps(fn)
    def wrapper(*args, **kwargs):
        if auth.is_enabled() and request.path.startswith('/api/') and not request.path.startswith('/api/settings') and not request.path.startswith('/api/kubernetes/dump'):
            @jwt_required
            def run():
                application_id = kwargs.get('application_id')
                if application_id is not None:
                    user_id = get_jwt_identity()
                    if not check_role(user_id, application_id, request.method):
                        raise ApplicationUserRoleException
                return fn(*args, **kwargs)
            return run()
        else:
            return fn(*args, **kwargs)
    return wrapper



# coding: utf-8


from flask_restplus import Resource, Namespace

from . import api


misc_info_namespace = Namespace('misc', description='Misc Endpoint.')


@misc_info_namespace.route('/settings')
class Settings(Resource):
    def get(self):
        result = {
            'auth': api.dashboard_config.IS_ACTIVATE_AUTH
        }
        return result



from flask_restplus import fields


class DatetimeToTimestamp(fields.Raw):
    def format(self, value):
        return value.timestamp()


decimal_suffixes = {
    'm': 1.0e-3,
    'k': 1.0e+3,
    'M': 1.0e+6,
    'G': 1.0e+9,
    'T': 1.0e+12,
    'P': 1.0e+15,
    'E': 1.0e+18,
}


def kubernetes_cpu_to_float(cpu_str: str):
    """ Convert the Kubernetes CPU value to float """
    suffix = cpu_str[-1]
    if suffix.isnumeric():
        return float(cpu_str)
    elif suffix in decimal_suffixes.keys():
        value = float(cpu_str[:-1])
        return value * decimal_suffixes[suffix]
    else:
        raise ValueError(f'Please check the input CPU value: `{cpu_str}`')



from functools import wraps
from flask_jwt_simple import get_jwt_identity
from flask_restplus import Resource, Namespace, fields, reqparse

from drucker_dashboard.auth import fetch_role, ApplicationUserRoleException
from drucker_dashboard.models import db, ApplicationUserRole, Role, User


admin_info_namespace = Namespace('admin', description='Admin Endpoint.')
success_or_not = admin_info_namespace.model('Success', {
    'status': fields.Boolean(
        required=True
    ),
    'message': fields.String(
        required=True
    )
})
user_info = admin_info_namespace.model('User', {
    'auth_id': fields.String(required=True),
    'user_name': fields.String(required=True)
})
role_info = admin_info_namespace.model('Role', {
    'role': fields.String(required=True),
    'user': fields.Nested(user_info, required=True)
})


@admin_info_namespace.route('/users')
class ApiApplicationUsers(Resource):
    @admin_info_namespace.marshal_list_with(user_info)
    def get(self):
        return User.query.all()


def check_owner_role(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        user_id = get_jwt_identity()
        application_id = kwargs.get('application_id')
        role = fetch_role(application_id, user_id)
        if role is not None and role == Role.owner:
            return fn(*args, **kwargs)
        else:
            raise ApplicationUserRoleException
    return wrapper


@admin_info_namespace.route('/<int:application_id>/acl')
class ApiApplicationIdACL(Resource):
    method_decorators = [check_owner_role]

    save_acl_parser = reqparse.RequestParser()
    save_acl_parser.add_argument('uid', type=str, required=True, location='form')
    save_acl_parser.add_argument('role', type=str, required=True, location='form')

    delete_acl_parser = reqparse.RequestParser()
    delete_acl_parser.add_argument('uid', type=str, required=True, location='form')

    @admin_info_namespace.marshal_list_with(role_info)
    def get(self, application_id):
        roles = ApplicationUserRole.query.filter_by(application_id=application_id).all()
        return roles

    @admin_info_namespace.marshal_with(success_or_not)
    @admin_info_namespace.expect(save_acl_parser)
    def post(self, application_id):
        args = self.save_acl_parser.parse_args()
        uid = args['uid']
        uobj = db.session.query(User).filter(User.auth_id == uid).one_or_none()
        if uobj is None:
            return {"status": False}, 400

        role = args['role']
        # if role is added by first user, they will be owner
        roles = db.session.query(ApplicationUserRole).filter(
            ApplicationUserRole.application_id == application_id).count()
        if roles == 0:
            sender_id = get_jwt_identity()
            if uobj.user_id != sender_id:
                ownerObj = ApplicationUserRole(
                    application_id=application_id,
                    user_id=sender_id,
                    role=Role.owner.name)
                db.session.add(ownerObj)
                db.session.flush()
            else:
                role = Role.owner.name
        try:
            roleObj = db.session.query(ApplicationUserRole).filter(
                ApplicationUserRole.application_id == application_id,
                ApplicationUserRole.user_id == uobj.user_id).one_or_none()
            if roleObj is None:
                roleObj = ApplicationUserRole(
                    application_id=application_id,
                    user_id=uobj.user_id,
                    role=role)
                db.session.add(roleObj)
                db.session.flush()
                db.session.commit()
                db.session.close()
                return {"status": True, "message": "Success."}, 200
            else:
                return {"status": False, "message": "Already exist."}, 400
        except Exception:
            return {"status": False}, 400

    @admin_info_namespace.marshal_with(success_or_not)
    @admin_info_namespace.expect(save_acl_parser)
    def patch(self, application_id):
        args = self.save_acl_parser.parse_args()
        uid = args['uid']
        uobj = db.session.query(User).filter(User.auth_id == uid).one_or_none()
        if uobj is None:
            return {"status": False}, 400
        roleObj = db.session.query(ApplicationUserRole).filter(
            ApplicationUserRole.application_id == application_id,
            ApplicationUserRole.user_id == uobj.user_id).one()
        roleObj.role = args['role']
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}, 200

    @admin_info_namespace.marshal_with(success_or_not)
    @admin_info_namespace.expect(delete_acl_parser)
    def delete(self, application_id):
        args = self.delete_acl_parser.parse_args()
        uid = args['uid']
        uobj = db.session.query(User).filter(User.auth_id == uid).one_or_none()
        if uobj is None:
            return {"status": False}, 400

        ApplicationUserRole.query.filter(
            ApplicationUserRole.application_id == application_id,
            ApplicationUserRole.user_id == uobj.user_id).delete()
        db.session.commit()
        db.session.close()
        return {"status": True, "message": "Success."}, 200



import uuid
import datetime

from flask_jwt_simple import get_jwt_identity
from flask_restplus import Namespace, fields, Resource, reqparse
from werkzeug.datastructures import FileStorage

from . import api
from drucker_dashboard import DruckerDashboardClient
from drucker_dashboard.models import db, Application, Service, Evaluation, EvaluationResult, ApplicationUserRole, Role, User
from drucker_dashboard.apis import DatetimeToTimestamp
from drucker_dashboard.utils import HashUtil


app_info_namespace = Namespace('applications', description='Application Endpoint.')
success_or_not = app_info_namespace.model('Success', {
    'status': fields.Boolean(
        required=True
    ),
    'message': fields.String(
        required=True
    )
})
app_info = app_info_namespace.model('Application', {
    'application_id': fields.Integer(
        readOnly=True,
        description='Application ID.'
    ),
    'application_name': fields.String(
        required=True,
        description='Application name.',
        example='drucker-sample'
    ),
    'kubernetes_id': fields.Integer(
        required=False,
        description='Kubernetes cluster ID.'
    ),
    'description': fields.String(
        required=False,
        description='Description.',
        example='This is a sample.'
    ),
    'register_date': DatetimeToTimestamp(
        readOnly=True,
        description='Register date.'
    ),
    'confirm_date': DatetimeToTimestamp(
        readOnly=True,
        description='Existance confirmation date.'
    )
})


@app_info_namespace.route('/')
class ApiApplications(Resource):
    add_worker_parser = reqparse.RequestParser()
    add_worker_parser.add_argument('host', type=str, required=True, location='form')
    add_worker_parser.add_argument('description', type=str, required=False, location='form')

    @app_info_namespace.marshal_list_with(app_info)
    def get(self):
        """get_applications"""
        if api.dashboard_config.IS_ACTIVATE_AUTH:
            user_id = get_jwt_identity()
            uobj = db.session.query(User).filter(User.user_id == user_id).one()
            # applications which don't have users are also accessible
            application_ids = db.session.query(ApplicationUserRole.application_id).distinct().all()
            ids = [application_id for application_id, in application_ids]
            applications = db.session.query(Application).filter(~Application.application_id.in_(ids)).all()
            return [assoc.application for assoc in uobj.applications] + applications
        return Application.query.all()

    @app_info_namespace.marshal_with(success_or_not)
    @app_info_namespace.expect(add_worker_parser)
    def post(self):
        """add_non-kube_worker"""
        args = self.add_worker_parser.parse_args()
        host = args['host']
        display_name = uuid.uuid4().hex
        description = args['description']

        drucker_dashboard_application = DruckerDashboardClient(logger=api.logger, host=host)
        service_info = drucker_dashboard_application.run_service_info()
        application_name = service_info["application_name"]
        service_name = service_info["service_name"]
        service_level = service_info["service_level"]

        aobj = db.session.query(Application).filter(
            Application.application_name == application_name,
            Application.kubernetes_id == None).one_or_none()
        if aobj is None:
            aobj = Application(application_name=application_name,
                               description=description)
            db.session.add(aobj)
            db.session.flush()
        if api.dashboard_config.IS_ACTIVATE_AUTH:
            user_id = get_jwt_identity()
            role = db.session.query(ApplicationUserRole).filter(
                ApplicationUserRole.application_id == aobj.application_id,
                ApplicationUserRole.user_id == user_id).one_or_none()
            if role is None:
                roleObj = ApplicationUserRole(
                    application_id=aobj.application_id,
                    user_id=user_id,
                    role=Role.owner.name)
                db.session.add(roleObj)
                db.session.flush()
        sobj = db.session.query(Service).filter(
            Service.service_name == service_name).one_or_none()
        if sobj is None:
            sobj = Service(application_id=aobj.application_id,
                           service_name=service_name,
                           display_name=display_name,
                           service_level=service_level,
                           host=host,
                           description=description)
            db.session.add(sobj)
            db.session.flush()
        response_body = {"status": True, "message": "Success."}
        db.session.commit()
        db.session.close()
        return response_body

@app_info_namespace.route('/<int:application_id>')
class ApiApplications(Resource):
    edit_app_info_parser = reqparse.RequestParser()
    edit_app_info_parser.add_argument('description', type=str, required=True, location='form')

    @app_info_namespace.marshal_with(app_info)
    def get(self, application_id:int):
        """get_application"""
        return Application.query.filter_by(application_id=application_id).first_or_404()

    @app_info_namespace.marshal_with(success_or_not)
    @app_info_namespace.expect(edit_app_info_parser)
    def patch(self, application_id:int):
        """update_application"""
        args = self.edit_app_info_parser.parse_args()
        description = args['description']
        res = db.session.query(Application).filter(Application.application_id == application_id).one()
        res.description = description
        response_body = {"status": True, "message": "Success."}
        db.session.commit()
        db.session.close()
        return response_body

@app_info_namespace.route('/<int:application_id>/evaluation')
class ApiEvaluation(Resource):
    upload_parser = reqparse.RequestParser()
    upload_parser.add_argument('file', location='files', type=FileStorage, required=True)

    @app_info_namespace.expect(upload_parser)
    def post(self, application_id:int):
        """update data to be evaluated"""
        args = self.upload_parser.parse_args()
        file = args['file']
        checksum = HashUtil.checksum(file)

        eobj = db.session.query(Evaluation).filter(
            Evaluation.application_id == application_id,
            Evaluation.checksum == checksum).one_or_none()
        if eobj is not None:
            return {"status": True, "evaluation_id": eobj.evaluation_id}

        eval_data_path = "eval-{0:%Y%m%d%H%M%S}.txt".format(datetime.datetime.utcnow())

        sobj = Service.query.filter_by(application_id=application_id).first_or_404()

        drucker_dashboard_application = DruckerDashboardClient(logger=api.logger, host=sobj.host)
        response_body = drucker_dashboard_application.run_upload_evaluation_data(file, eval_data_path)

        if not response_body['status']:
            raise Exception('Failed to upload')
        eobj = Evaluation(checksum=checksum, application_id=application_id, data_path=eval_data_path)
        db.session.add(eobj)
        db.session.flush()
        evaluation_id = eobj.evaluation_id
        db.session.commit()
        db.session.close()

        return {"status": True, "evaluation_id": evaluation_id}


@app_info_namespace.route('/<int:application_id>/evaluation/<int:evaluation_id>')
class ApiEvaluation(Resource):
    def delete(self, application_id:int, evaluation_id:int):
        """delete data to be evaluated"""
        eval_query = db.session.query(Evaluation)\
            .filter(Evaluation.application_id == application_id,
                    Evaluation.evaluation_id == evaluation_id)
        if eval_query.one_or_none() is None:
            return {"status": False}, 404

        eval_query.delete()
        db.session.query(EvaluationResult)\
            .filter(EvaluationResult.evaluation_id == evaluation_id).delete()
        db.session.commit()
        db.session.close()

        return {"status": True, "message": "Success."}



import datetime
from .dao import db

from sqlalchemy import (
    Column, Integer, String,
    Text, DateTime, UniqueConstraint
)


class Application(db.Model):
    """
    Applications
    """
    __tablename__ = 'applications'
    __table_args__ = (
        UniqueConstraint('application_id'),
        {'mysql_engine': 'InnoDB'}
    )

    application_id = Column(Integer, primary_key=True, autoincrement=True)
    application_name = Column(String(128), nullable=False)
    kubernetes_id = Column(Integer, nullable=True)
    description = Column(Text, nullable=True)
    register_date = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)
    confirm_date = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)

    @property
    def serialize(self):
        return {
            'application_id': self.application_id,
            'application_name': self.application_name,
            'kubernetes_id': self.kubernetes_id,
            'description': self.description,
            'register_date': self.register_date.strftime('%Y-%m-%d %H:%M:%S'),
            'confirm_date': self.confirm_date.strftime('%Y-%m-%d %H:%M:%S'),
        }



# coding: utf-8


import yaml
import os
import json
from drucker_dashboard.protobuf import drucker_pb2

from .hash_util import HashUtil
from .protobuf_util import ProtobufUtil


class DruckerDashboardConfig:
    def __init__(self, config_file: str):
        settings_yaml = os.getenv("DRUCKER_SETTINGS_YAML", config_file)
        config = dict()
        if settings_yaml is not None:
            with open(settings_yaml, 'r') as f:
                config = yaml.load(f)
        self.TEST_MODE = str(os.getenv("DRUCKER_TEST_MODE", config.get("test", "False"))).lower() == 'true'
        self.DRUCKER_GRPC_VERSION = drucker_pb2.DESCRIPTOR.GetOptions().Extensions[drucker_pb2.drucker_grpc_proto_version]
        self.DIR_KUBE_CONFIG = os.getenv('DRUCKER_KUBE_DATADIR', config.get('kube.datadir', 'kube-config'))
        self._DB_MODE = os.getenv('DRUCKER_DB_MODE', config.get('use.db',"sqlite"))
        self._DB_MYSQL_HOST = os.getenv('DRUCKER_DB_MYSQL_HOST', config.get('db.mysql.host',""))
        self._DB_MYSQL_PORT = os.getenv('DRUCKER_DB_MYSQL_PORT', config.get('db.mysql.port',""))
        self._DB_MYSQL_DBNAME = os.getenv('DRUCKER_DB_MYSQL_DBNAME', config.get('db.mysql.dbname',""))
        self._DB_MYSQL_USER = os.getenv('DRUCKER_DB_MYSQL_USER', config.get('db.mysql.user',""))
        self._DB_MYSQL_PASSWORD = os.getenv('DRUCKER_DB_MYSQL_PASSWORD', config.get('db.mysql.password',""))

        if self._DB_MODE == "sqlite":
            db_name = "db.test.sqlite3" if self.TEST_MODE else "db.sqlite3"
            self.DB_URL = f'sqlite:///{db_name}'
        elif self._DB_MODE == "mysql":
            host = self._DB_MYSQL_HOST
            port = self._DB_MYSQL_PORT
            db_name = "test_"+self._DB_MYSQL_DBNAME if self.TEST_MODE else self._DB_MYSQL_DBNAME
            user = self._DB_MYSQL_USER
            password = self._DB_MYSQL_PASSWORD
            self.DB_URL = f'mysql+pymysql://{user}:{password}@{host}:{port}/{db_name}?charset=utf8'
        else:
            self.DB_URL = 'sqlite:///db.test.sqlite3'

        if str(os.getenv('DRUCKER_ACTIVATE_AUTH', 'False')).lower() == 'true':
            self.IS_ACTIVATE_AUTH = True
            self.AUTH_CONFIG = {
                'secret': os.getenv('DRUCKER_LDAP_SECRET'),
                'ldap': {
                    'host': os.getenv('DRUCKER_LDAP_HOST'),
                    'port': os.getenv('DRUCKER_LDAP_PORT'),
                    'bind_dn': os.getenv('DRUCKER_LDAP_BIND_DN'),
                    'bind_password': os.getenv('DRUCKER_LDAP_BIND_PASSWORD'),
                    'search_filter': os.getenv('DRUCKER_LDAP_SEARCH_FILTER'),
                    'search_base_dns': json.loads(os.getenv('DRUCKER_LDAP_SEARCH_BASE_DNS')),
                }
            }
        elif 'auth' in config:
            self.IS_ACTIVATE_AUTH = True
            self.AUTH_CONFIG = config['auth']
        else:
            self.IS_ACTIVATE_AUTH = False
            self.AUTH_CONFIG = None



import datetime
from .dao import db
from sqlalchemy import (
    Column, Integer, DateTime,
    String, UniqueConstraint
)
from sqlalchemy.orm import relationship


class Evaluation(db.Model):
    """
    Data to be evaluated
    """
    __tablename__ = 'evaluations'
    __table_args__ = (
        UniqueConstraint('evaluation_id'),
        UniqueConstraint('checksum'),
        {'mysql_engine': 'InnoDB'}
    )

    evaluation_id = Column(Integer, primary_key=True, autoincrement=True)
    checksum = Column(String(128), nullable=False)
    application_id = Column(Integer, nullable=False)
    data_path = Column(String(512), nullable=False)
    register_date = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)
    evaluation_result = relationship('EvaluationResult', backref='evaluations', lazy='select', innerjoin=True)

    @property
    def serialize(self):
        return {
            'evaluation_id': self.evaluation_id,
            'checksum': self.checksum,
            'application_id': self.application_id,
            'data_path': self.data_path,
            'register_date': self.register_date.strftime('%Y-%m-%d %H:%M:%S'),
        }



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# DO NOT EDIT HERE!!

import traceback, types
import grpc

from .protobuf import drucker_pb2, drucker_pb2_grpc

from .logger.logger_interface import SystemLoggerInterface
from .utils.protobuf_util import ProtobufUtil
from werkzeug.datastructures import FileStorage
from protobuf_to_dict import protobuf_to_dict


def error_handling(error_response):
    """ Decorator for handling error

    Apply following processing on Servicer methods
    to handle errors.

    - Call :func:``on_error`` method (if defined) in the class
      to postprocess something on error

    Parameters
    ----------
    error_response
    """
    def _wrapper_maker(func):
        def _wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as error:
                servicer = args[0]
                if hasattr(servicer, 'on_error'):
                    assert isinstance(servicer.on_error, types.MethodType), \
                        'You must define on_error as method'
                    servicer.on_error(error)
                return error_response
        return _wrapper
    return _wrapper_maker


class DruckerDashboardClient:
    def __init__(self, logger:SystemLoggerInterface, host:str=None, domain:str=None, app:str=None, env:str=None, version:int=None):
        self.logger = logger
        self.stub = None
        if host is None and (domain is None or app is None or env is None):
            raise RuntimeError("You must specify host or domain+app+env.")

        if version is None:
            v_str = drucker_pb2.DESCRIPTOR.GetOptions().Extensions[drucker_pb2.drucker_grpc_proto_version]
        else:
            v_str = drucker_pb2.EnumVersionInfo.Name(version)

        if host is None:
            self.__change_domain_app_env(domain, app, env, v_str)
        else:
            self.__change_host(host)

    def on_error(self, error: Exception):
        """ Postprocessing on error

        For detail, see :func:``on_error``

        Parameters
        ----------
        error : Exception
            Error to be handled
        """
        self.logger.error(str(error))
        self.logger.error(traceback.format_exc())

    def __change_domain_app_env(self, domain:str, app:str, env:str, version:str):
        host = "{0}-{1}-{2}.{3}".format(app,version,env,domain)
        self.__change_host(host)

    def __change_host(self, host:str):
        channel = grpc.insecure_channel(host)
        self.stub = drucker_pb2_grpc.DruckerDashboardStub(channel)

    @error_handling({"status": False})
    def run_service_info(self):
        request = drucker_pb2.ServiceInfoRequest()
        response_protobuf = self.stub.ServiceInfo(request)
        response = protobuf_to_dict(response_protobuf,
                                    including_default_value_fields=True)
        return response

    def __upload_model_request(self, model_path:str, f:FileStorage):
        for data in ProtobufUtil.stream_file(f):
            request = drucker_pb2.UploadModelRequest(
                path=model_path,
                data=data
            )
            yield request

    @error_handling({"status": False})
    def run_upload_model(self, model_path:str, f:FileStorage):
        request_iterator = self.__upload_model_request(model_path, f)
        response = self.stub.UploadModel(request_iterator)
        response = protobuf_to_dict(response,
                                    including_default_value_fields=True)
        if not response["status"]:
            raise Exception(response["message"])
        return response

    @error_handling({"status": False})
    def run_switch_service_model_assignment(self, model_path:str):
        request = drucker_pb2.SwitchModelRequest(
            path=model_path,
        )
        response_protobuf = self.stub.SwitchModel(request)
        response = protobuf_to_dict(response_protobuf,
                                    including_default_value_fields=True)
        if not response["status"]:
            raise Exception(response["message"])
        return response

    @error_handling({"status": False})
    def run_evaluate_model(self, data_path:str, result_path:str):
        request_iterator = iter([drucker_pb2.EvaluateModelRequest(data_path=data_path, result_path=result_path)])
        response = protobuf_to_dict(self.stub.EvaluateModel(request_iterator).metrics,
                                    including_default_value_fields=True)
        response['status'] = True
        return response

    def __upload_eval_data_request(self, f:FileStorage, data_path:str):
        for data in ProtobufUtil.stream_file(f):
            request = drucker_pb2.UploadEvaluationDataRequest(data=data, data_path=data_path)
            yield request

    @error_handling({"status": False})
    def run_upload_evaluation_data(self, f:FileStorage, data_path:str):
        request_iterator = self.__upload_eval_data_request(f, data_path)
        response = protobuf_to_dict(self.stub.UploadEvaluationData(request_iterator),
                                    including_default_value_fields=True)
        return response



# -*- coding: utf-8 -*-


import argparse

from drucker_dashboard import _version
from drucker_dashboard.server import create_app


def server_handler(args):
    app = create_app(args.settings, args.logger)
    app.run(host=args.host, port=args.port, threaded=True)


def db_handler(args):
    import sys
    from flask_script import Manager
    from flask_migrate import Migrate, MigrateCommand
    from drucker_dashboard.models import db

    tmp = sys.argv[1:]
    sys.argv = [sys.argv[0]]
    for i, t in enumerate(tmp):
        if t == 'db':
            sys.argv.append(t)
            sys.argv.append(tmp[i+1])
            break
    app = create_app(args.settings, args.logger)
    migrate = Migrate(app, db)
    manager = Manager(app)
    manager.add_command('db', MigrateCommand)
    manager.run()


def create_parser():
    parser = argparse.ArgumentParser(description='rekcurdui command')
    parser.add_argument(
        '--version', '-v', action='version', version=_version.__version__)
    parser.add_argument(
        '--settings', required=True, help='settings YAML. See https://github.com/rekcurd/drucker-dashboard/blob/master/drucker_dashboard/settings.yml')
    parser.add_argument(
        '--logger', required=False, help='Python file of your custom logger. Need to inherit "logger_interface.py". See https://github.com/rekcurd/drucker-dashboard/blob/master/drucker_dashboard/logger/logger_interface.py', default=None)
    subparsers = parser.add_subparsers()

    # server
    parser_server = subparsers.add_parser(
        'server', help='see `rekcurdui server -h`')
    parser_server.add_argument(
        '-H', '--host', required=False, help='host', default='0.0.0.0')
    parser_server.add_argument(
        '-p', '--port', required=False, type=int, help='port', default=18080)
    parser_server.set_defaults(handler=server_handler)

    # db
    parser_db = subparsers.add_parser(
        'db', help='see `rekcurdui db -h`')
    parser_db.add_argument(
        'type', choices=['init', 'revision', 'migrate', 'edit', 'merge',
                         'upgrade', 'downgrade', 'show', 'history', 'heads',
                         'branches', 'current', 'stamp'])
    parser_db.set_defaults(handler=db_handler)

    return parser


def main() -> None:
    parser = create_parser()
    args = parser.parse_args()

    if hasattr(args, 'handler'):
        args.handler(args)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



import datetime
from .dao import db
from sqlalchemy import (
    Column, Integer, String, DateTime,
    Text, UniqueConstraint
)


class Kubernetes(db.Model):
    """
    Kubernetes Info
    """
    __tablename__ = 'kubernetes'
    __table_args__ = (
        UniqueConstraint('kubernetes_id'),
        UniqueConstraint('config_path'),
        UniqueConstraint('dns_name'),
        UniqueConstraint('display_name'),
        {'mysql_engine': 'InnoDB'}
    )

    kubernetes_id = Column(Integer, primary_key=True, autoincrement=True)
    config_path = Column(String(512), nullable=False)
    dns_name = Column(String(512), nullable=False)
    display_name = Column(String(128), nullable=False)
    db_mysql_host = Column(String(128), nullable=False)
    db_mysql_port = Column(String(128), nullable=False)
    db_mysql_dbname = Column(String(128), nullable=False)
    db_mysql_user = Column(String(128), nullable=False)
    db_mysql_password = Column(String(128), nullable=False)
    host_model_dir = Column(String(128), nullable=False)
    pod_model_dir = Column(String(128), nullable=False)
    register_date = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)
    description = Column(Text, nullable=True)

    @property
    def serialize(self):
        return {
            'kubernetes_id': self.kubernetes_id,
            'config_path': self.config_path,
            'dns_name': self.dns_name,
            'display_name': self.display_name,
            'db_mysql_host': self.db_mysql_host,
            'db_mysql_port': self.db_mysql_port,
            'db_mysql_dbname': self.db_mysql_dbname,
            'db_mysql_user': self.db_mysql_user,
            'db_mysql_password': self.db_mysql_password,
            'host_model_dir': self.host_model_dir,
            'pod_model_dir': self.pod_model_dir,
            'register_date': self.register_date.strftime('%Y-%m-%d %H:%M:%S'),
            'description': self.description,
        }



import os, uuid, shutil, json, pathlib
from datetime import datetime, timedelta

from flask_restplus import Namespace, fields, Resource, reqparse

from werkzeug.datastructures import FileStorage

from . import api, DatetimeToTimestamp, kubernetes_cpu_to_float
from drucker_dashboard import DruckerDashboardClient
from drucker_dashboard.models import db, Kubernetes, Application, Service, Model


kube_info_namespace = Namespace('kubernetes', description='Kubernetes Endpoint.')
success_or_not = kube_info_namespace.model('Success', {
    'status': fields.Boolean(
        required=True
    ),
    'message': fields.String(
        required=True
    )
})
kube_info = kube_info_namespace.model('Kubernetes', {
    'kubernetes_id': fields.Integer(
        readOnly=True,
        description='Kubernetes cluster ID.'
    ),
    'config_path': fields.String(
        readOnly=True,
        description='Kubernetes configuration file path.',
        example='/kube-config/kube-1.config'
    ),
    'dns_name': fields.String(
        readOnly=False,
        description='DNS name.',
        example='example.com'
    ),
    'display_name': fields.String(
        required=False,
        description='DisplayName.',
        example='cluster-1'
    ),
    'host_model_dir': fields.String(
        required=True,
        description='Directry of host node for ML models. Recommend that mount cloud storage (e.g. AWS EBS, GCP GCS) on <host_model_dir>.',
        example='/mnt/drucker-model'
    ),
    'pod_model_dir': fields.String(
        required=True,
        description='Directry of pod for ML models. Link to <host_model_dir>.',
        example='/mnt/drucker-model'
    ),
    'db_mysql_host': fields.String(
        required=True,
        description='mysql host DNS/IP.',
        example='127.0.0.1'
    ),
    'db_mysql_port': fields.String(
        required=True,
        description='mysql port.',
        example='2306'
    ),
    'db_mysql_dbname': fields.String(
        required=True,
        description='mysql DB name.',
        example='drucker'
    ),
    'db_mysql_user': fields.String(
        required=True,
        description='mysql user.',
        example='drucker'
    ),
    'db_mysql_password': fields.String(
        required=True,
        description='mysql password.',
        example='drucker'
    ),
    'register_date': DatetimeToTimestamp(
        readOnly=True,
        description='Register date.'
    ),
    'description': fields.String(
        required=False,
        description='Description.',
        example='This is a sample.'
    )
})
kube_service_config_info = kube_info_namespace.model('Kubernetes service configurations', {
    'app_name': fields.String(
        required=True,
        description='Application name.',
        example='drucker-sample',
        max_length=20
    ),
    'service_level': fields.String(
        required=True,
        description='Service level. Choose from [development/beta/staging/sandbox/production].',
        example='development'
    ),
    'service_name': fields.String(
        readOnly=True,
        description='Service tag.',
        example='dev-123456789',
        max_length=20
    ),
    'service_port': fields.Integer(
        required=True,
        description='Service port.',
        example=5000
    ),
    'replicas_default': fields.Integer(
        required=True,
        description='Number of first replica.',
        example=1
    ),
    'replicas_minimum': fields.Integer(
        required=True,
        description='Minimum number of replica for auto-scaling.',
        example=1
    ),
    'replicas_maximum': fields.Integer(
        required=True,
        description='Maximum number of replica for auto-scaling.',
        example=1
    ),
    'autoscale_cpu_threshold': fields.Integer(
        required=True,
        description='CPU usage threshold for auto-scaling.',
        example=80
    ),
    'policy_max_surge': fields.Integer(
        required=True,
        description='Maximum number of serged pod when updating. Recommendation is "ceil(0.25 * <replicas_default>)".',
        example=1
    ),
    'policy_max_unavailable': fields.Integer(
        required=True,
        description='Maximum number of unavailable pod when updating. Recommendation is "floor(0.25 * <replicas_default>)".',
        example=0
    ),
    'policy_wait_seconds': fields.Integer(
        required=True,
        description='Minimum wait seconds when updating. Recommendation is "actual boot time + alpha [second]".',
        example=300
    ),
    'container_image': fields.String(
        required=True,
        description='Image location of Docker registry.',
        example='centos:centos7'
    ),
    'resource_request_cpu': fields.Float(
        required=True,
        description='CPU request for boot.',
        example=1.0
    ),
    'resource_request_memory': fields.String(
        required=True,
        description='Memory request for boot.',
        example='128Mi'
    ),
    'resource_limit_cpu': fields.Float(
        required=True,
        description='CPU limit for boot.',
        example=1.0
    ),
    'resource_limit_memory': fields.String(
        required=True,
        description='Memory limit for boot.',
        example='128Mi'
    ),
    'commit_message': fields.String(
        required=False,
        description='Commit message.',
        example='Initial deployment.'
    ),
    'service_git_url': fields.String(
        required=True,
        description='Git repository.',
        example='git@github.com:drucker/drucker-example.git'
    ),
    'service_git_branch': fields.String(
        required=True,
        description='Git branch.',
        example='master'
    ),
    'service_boot_script': fields.String(
        required=True,
        description='Boot shellscript for your service.',
        example='start.sh'
    ),
    'host_model_dir': fields.String(
        required=True,
        description='Directry of host node for ML models. Recommend that mount cloud storage (e.g. AWS EBS, GCP GCS) on <host_model_dir>.',
        example='/mnt/drucker-model'
    ),
    'pod_model_dir': fields.String(
        required=True,
        description='Directry of pod for ML models. Link to <host_model_dir>.',
        example='/mnt/drucker-model'
    ),
    'db_mysql_host': fields.String(
        required=True,
        description='mysql host DNS/IP.',
        example='127.0.0.1'
    ),
    'db_mysql_port': fields.String(
        required=True,
        description='mysql port.',
        example='2306'
    ),
    'db_mysql_dbname': fields.String(
        required=True,
        description='mysql DB name.',
        example='drucker'
    ),
    'db_mysql_user': fields.String(
        required=True,
        description='mysql user.',
        example='drucker'
    ),
    'db_mysql_password': fields.String(
        required=True,
        description='mysql password.',
        example='drucker'
    )
})

kube_file_parser = reqparse.RequestParser()
kube_file_parser.add_argument('file', location='files', type=FileStorage, required=True, help='Kubernetes access_token configuration file.')
kube_file_parser.add_argument('dns_name', type=str, default='example.com', required=True, help='Your application could be accessed at http://<app_name>-<service_level>.<dns_name>.', location='form')
kube_file_parser.add_argument('display_name', type=str, required=False, location='form', help="Must be unique. If empty, automatically generated.")
kube_file_parser.add_argument('host_model_dir', type=str, default='/mnt/drucker-model', required=True, help='Directry of host node for ML models. Recommend that mount cloud storage (e.g. AWS EBS, GCP GCS) on <host_model_dir>.', location='form')
kube_file_parser.add_argument('pod_model_dir', type=str, default='/mnt/drucker-model', required=True, help='Directry of pod for ML models. Link to <host_model_dir>.', location='form')
kube_file_parser.add_argument('db_mysql_host', type=str, default='127.0.0.1', required=True, help='mysql host DNS/IP.', location='form')
kube_file_parser.add_argument('db_mysql_port', type=str, default='2306', required=True, help='mysql port.', location='form')
kube_file_parser.add_argument('db_mysql_dbname', type=str, default='drucker', required=True, help='mysql DB name.', location='form')
kube_file_parser.add_argument('db_mysql_user', type=str, default='drucker', required=True, help='mysql user.', location='form')
kube_file_parser.add_argument('db_mysql_password', type=str, default='drucker', required=True, help='mysql password.', location='form')
kube_file_parser.add_argument('description', type=str, required=False, location='form', help='Description.')

kube_deploy_parser = reqparse.RequestParser()
kube_deploy_parser.add_argument('app_name', type=str, default='drucker-sample', required=True, help='Application name. This must be unique.', location='form')
kube_deploy_parser.add_argument('service_level', type=str, required=True, choices=('development','beta','staging','sandbox','production'), help='Service level. Choose from [development/beta/staging/sandbox/production].', location='form')
kube_deploy_parser.add_argument('service_port', type=int, default=5000, required=True, help='Service port.', location='form')

kube_deploy_parser.add_argument('replicas_default', type=int, default=1, required=True, help='Number of first replica.', location='form')
kube_deploy_parser.add_argument('replicas_minimum', type=int, default=1, required=True, help='Minimum number of replica for auto-scaling.', location='form')
kube_deploy_parser.add_argument('replicas_maximum', type=int, default=1, required=True, help='Maximum number of replica for auto-scaling.', location='form')
kube_deploy_parser.add_argument('autoscale_cpu_threshold', type=int, default=80, required=True, help='CPU usage threshold for auto-scaling.', location='form')

kube_deploy_parser.add_argument('policy_max_surge', type=int, default=1, required=True, help='Maximum number of serged pod when updating. Recommendation is "ceil(0.25 * <replicas_default>)".', location='form')
kube_deploy_parser.add_argument('policy_max_unavailable', type=int, default=0, required=True, help='Maximum number of unavailable pod when updating. Recommendation is "floor(0.25 * <replicas_default>)".', location='form')
kube_deploy_parser.add_argument('policy_wait_seconds', type=int, default=300, required=True, help='Minimum wait seconds when updating. Recommendation is "actual boot time + alpha [second]"', location='form')

kube_deploy_parser.add_argument('container_image', type=str, default='centos:centos7', required=True, help='Image location of Docker registry.', location='form')

kube_deploy_parser.add_argument('resource_request_cpu', type=float, default=1.0, required=True, help='CPU request for boot.', location='form')
kube_deploy_parser.add_argument('resource_request_memory', type=str, default='128Mi', required=True, help='Memory request for boot.', location='form')
kube_deploy_parser.add_argument('resource_limit_cpu', type=float, default=1.0, required=True, help='CPU limit for boot.', location='form')
kube_deploy_parser.add_argument('resource_limit_memory', type=str, default='256Mi', required=True, help='Memory limit for boot.', location='form')

kube_deploy_parser.add_argument('commit_message', type=str, default='Initial deployment.', required=False, help='Commit message.', location='form')
kube_deploy_parser.add_argument('service_git_url', type=str, default='git@github.com:drucker/drucker-sample.git', required=True, help='Git repository.', location='form')
kube_deploy_parser.add_argument('service_git_branch', type=str, default='master', required=True, help='Git branch.', location='form')
kube_deploy_parser.add_argument('service_boot_script', type=str, default='start.sh', required=True, help='Boot shellscript for your service.', location='form')

kube_deploy_parser.add_argument('service_model_assignment', type=int, required=False, help='Model assignment when service boots.', location='form')


def update_dbs_kubernetes(kubernetes_id:int, applist:set=None, description:str=None):
    """
    Update dbs of kubernetes entry.
    :param kubernetes_id:
    :return:
    """
    process_date = datetime.utcnow() - timedelta(seconds=1)
    kobj = db.session.query(Kubernetes).filter(
        Kubernetes.kubernetes_id == kubernetes_id).one_or_none()
    if kobj is None:
        raise Exception("No such kubernetes_id.")
    config_path = kobj.config_path
    from kubernetes import client, config
    config.load_kube_config(config_path)
    v1 = client.ExtensionsV1beta1Api()
    ret = v1.list_ingress_for_all_namespaces(watch=False)

    refresh_all = False
    if applist is None:
        refresh_all = True
        applist = set()
        for i in ret.items:
            labels = i.metadata.labels
            if labels is not None and labels.get("drucker-worker", "False") == "True":
                application_name = labels["app"]
                applist.add(application_name)
    """Application"""
    for application_name in applist:
        aobj = db.session.query(Application).filter(
            Application.application_name == application_name,
            Application.kubernetes_id == kubernetes_id).one_or_none()
        if aobj is None:
            aobj = Application(application_name=application_name,
                               kubernetes_id=kubernetes_id,
                               description=description)
            db.session.add(aobj)
            db.session.flush()
        else:
            aobj.confirm_date = datetime.utcnow()
            db.session.flush()
    """Service"""
    for i in ret.items:
        labels = i.metadata.labels
        if labels is None or labels.get("drucker-worker", "False") == "False":
            continue
        application_name = labels["app"]
        service_name = labels["sel"]
        service_level = i.metadata.namespace
        host = i.spec.rules[0].host
        aobj = db.session.query(Application).filter(
            Application.application_name == application_name,
            Application.kubernetes_id == kubernetes_id).one_or_none()
        if aobj.application_name not in applist:
            continue

        sobj = db.session.query(Service).filter(
            Service.service_name == service_name).one_or_none()
        if sobj is None:
            display_name = uuid.uuid4().hex
            sobj = Service(application_id=aobj.application_id,
                           service_name=service_name,
                           service_level=service_level,
                           display_name=display_name,
                           host=host,
                           description=description)
            db.session.add(sobj)
            db.session.flush()
            sobj.display_name = "{0}-{1}".format(service_level,sobj.service_id)
            db.session.flush()
        else:
            sobj.confirm_date = datetime.utcnow()
            db.session.flush()
    for application_name in applist:
        aobj = db.session.query(Application).filter(
            Application.application_name == application_name,
            Application.kubernetes_id == kubernetes_id).one_or_none()
        db.session.query(Service).filter(
            Service.confirm_date <= process_date,
            Service.application_id == aobj.application_id).delete()
        db.session.flush()
    if refresh_all:
        db.session.query(Application).filter(
            Application.confirm_date <= process_date,
            Application.kubernetes_id == kubernetes_id).delete()
        db.session.flush()


def create_or_update_drucker_on_kubernetes(
        kubernetes_id:int, args:dict, service_name:str=None):
    app_name = args['app_name']
    if len(app_name) > 20:
        raise Exception("Application Name is too long. Up to 20.")
    service_level = args['service_level']
    service_port = args['service_port']
    replicas_default = args['replicas_default']
    replicas_minimum = args['replicas_minimum']
    replicas_maximum = args['replicas_maximum']
    autoscale_cpu_threshold = args['autoscale_cpu_threshold']
    policy_max_surge = args['policy_max_surge']
    policy_max_unavailable = args['policy_max_unavailable']
    policy_wait_seconds = args['policy_wait_seconds']
    container_image = args['container_image']
    resource_request_cpu = args['resource_request_cpu']
    resource_request_memory = args['resource_request_memory']
    resource_limit_cpu = args['resource_limit_cpu']
    resource_limit_memory = args['resource_limit_memory']
    commit_message = args['commit_message']
    service_git_url = args['service_git_url']
    service_git_branch = args['service_git_branch']
    service_boot_script = args['service_boot_script']
    service_model_assignment = args['service_model_assignment']

    mode_create = False
    if service_name is None:
        mode_create = True
        service_name = "{0}-{1}-{2:%Y%m%d%H%M%S}".format(
            app_name,service_level,datetime.utcnow())
    volume_name = "host-volume"
    ssh_volume = "ssh-volume"
    ssh_dir = "/root/.ssh"
    num_retry = 5
    progress_deadline_seconds = int(num_retry*policy_wait_seconds*replicas_maximum/(policy_max_surge+policy_max_unavailable))

    kobj = db.session.query(Kubernetes).filter(
        Kubernetes.kubernetes_id == kubernetes_id).one_or_none()
    if kobj is None:
        raise Exception("No such kubernetes_id.")
    app_dns_name = kobj.dns_name
    host_model_dir = kobj.host_model_dir
    pod_model_dir = kobj.pod_model_dir
    db_mysql_host = kobj.db_mysql_host
    db_mysql_port = kobj.db_mysql_port
    db_mysql_dbname = kobj.db_mysql_dbname
    db_mysql_user = kobj.db_mysql_user
    db_mysql_password = kobj.db_mysql_password
    config_path = kobj.config_path
    from kubernetes import client, config
    config.load_kube_config(config_path)

    pod_env = [
        client.V1EnvVar(
            name="DRUCKER_SERVICE_UPDATE_FLAG",
            value=commit_message
        ),
        client.V1EnvVar(
            name="DRUCKER_TEST_MODE",
            value="False"
        ),
        client.V1EnvVar(
            name="DRUCKER_APPLICATION_NAME",
            value=app_name
        ),
        client.V1EnvVar(
            name="DRUCKER_SERVICE_LEVEL",
            value=service_level
        ),
        client.V1EnvVar(
            name="DRUCKER_SERVICE_NAME",
            value=service_name
        ),
        client.V1EnvVar(
            name="DRUCKER_SERVICE_PORT",
            value="{0}".format(service_port)
        ),
        client.V1EnvVar(
            name="DRUCKER_SERVICE_INFRA",
            value="kubernetes"
        ),
        client.V1EnvVar(
            name="DRUCKER_SERVICE_GIT_URL",
            value=service_git_url
        ),
        client.V1EnvVar(
            name="DRUCKER_SERVICE_GIT_BRANCH",
            value=service_git_branch
        ),
        client.V1EnvVar(
            name="DRUCKER_SERVICE_BOOT_SHELL",
            value=service_boot_script
        ),
        client.V1EnvVar(
            name="DRUCKER_SERVICE_MODEL_DIR",
            value=pod_model_dir
        ),
        client.V1EnvVar(
            name="DRUCKER_DB_MODE",
            value="mysql"
        ),
        client.V1EnvVar(
            name="DRUCKER_DB_MYSQL_HOST",
            value=db_mysql_host
        ),
        client.V1EnvVar(
            name="DRUCKER_DB_MYSQL_PORT",
            value=db_mysql_port
        ),
        client.V1EnvVar(
            name="DRUCKER_DB_MYSQL_DBNAME",
            value=db_mysql_dbname
        ),
        client.V1EnvVar(
            name="DRUCKER_DB_MYSQL_USER",
            value=db_mysql_user
        ),
        client.V1EnvVar(
            name="DRUCKER_DB_MYSQL_PASSWORD",
            value=db_mysql_password
        ),
    ]
    if service_model_assignment is not None:
        mobj = Model.query.filter_by(
            model_id=service_model_assignment).one()
        pod_env.append(
            client.V1EnvVar(
                name="DRUCKER_SERVICE_MODEL_FILE",
                value=mobj.model_path
            )
        )

    """Namespace"""
    core_vi = client.CoreV1Api()
    try:
        api_response = core_vi.read_namespace(name=service_level)
    except:
        v1_namespace = client.V1Namespace(
            api_version="v1",
            kind="Namespace",
            metadata=client.V1ObjectMeta(
                name=service_level
            )
        )
        api_response = core_vi.create_namespace(
            body=v1_namespace
        )
    """Deployment"""
    v1_deployment = client.V1Deployment(
        api_version="apps/v1",
        kind="Deployment",
        metadata=client.V1ObjectMeta(
            name="{0}-deployment".format(service_name),
            namespace=service_level,
            labels={"drucker-worker": "True", "app": app_name, "sel": service_name}
        ),
        spec=client.V1DeploymentSpec(
            min_ready_seconds=policy_wait_seconds,
            progress_deadline_seconds=progress_deadline_seconds,
            replicas=replicas_default,
            revision_history_limit=3,
            selector=client.V1LabelSelector(
                match_labels={"sel": service_name}
            ),
            strategy=client.V1DeploymentStrategy(
                type="RollingUpdate",
                rolling_update=client.V1RollingUpdateDeployment(
                    max_surge=policy_max_surge,
                    max_unavailable=policy_max_unavailable)
            ),
            template=client.V1PodTemplateSpec(
                metadata=client.V1ObjectMeta(
                    labels={"drucker-worker": "True", "app": app_name, "sel": service_name}
                ),
                spec=client.V1PodSpec(
                    affinity=client.V1Affinity(
                        pod_anti_affinity=client.V1PodAntiAffinity(
                            preferred_during_scheduling_ignored_during_execution=[
                                client.V1WeightedPodAffinityTerm(
                                    pod_affinity_term=client.V1PodAffinityTerm(
                                        label_selector=client.V1LabelSelector(
                                            match_expressions=[
                                                client.V1LabelSelectorRequirement(
                                                    key="app",
                                                    operator="In",
                                                    values=[service_name]
                                                )
                                            ]
                                        ),
                                        topology_key="kubernetes.io/hostname"
                                    ),
                                    weight=100
                                )
                            ]
                        )
                    ),
                    containers=[
                        client.V1Container(
                            command=["sh","/usr/local/src/entrypoint.sh"],
                            env=pod_env,
                            image=container_image,
                            image_pull_policy="Always",
                            name=service_name,
                            ports=[
                                client.V1ContainerPort(container_port=service_port)
                            ],
                            resources=client.V1ResourceRequirements(
                                limits={
                                    "cpu": "{0}".format(resource_limit_cpu),
                                    "memory": resource_limit_memory
                                },
                                requests={
                                    "cpu": "{0}".format(resource_request_cpu),
                                    "memory": resource_request_memory
                                }
                            ),
                            security_context=client.V1SecurityContext(
                                privileged=True
                            ),
                            volume_mounts=[
                                client.V1VolumeMount(
                                    mount_path=pod_model_dir,
                                    name=volume_name
                                ),
                                client.V1VolumeMount(
                                    mount_path=ssh_dir,
                                    name=ssh_volume
                                )
                            ]
                        )
                    ],
                    node_selector={"host": service_level},
                    volumes=[
                        client.V1Volume(
                            host_path=client.V1HostPathVolumeSource(
                                path=host_model_dir
                            ),
                            name=volume_name
                        ),
                        client.V1Volume(
                            host_path=client.V1HostPathVolumeSource(
                                path=ssh_dir
                            ),
                            name=ssh_volume
                        )
                    ]
                )
            )
        )
    )
    apps_v1 = client.AppsV1Api()
    if mode_create:
        api_response = apps_v1.create_namespaced_deployment(
            body=v1_deployment,
            namespace=service_level
        )
    else:
        api_response = apps_v1.patch_namespaced_deployment(
            body=v1_deployment,
            name="{0}-deployment".format(service_name),
            namespace=service_level
        )
    """Service"""
    v1_service = client.V1Service(
        api_version="v1",
        kind="Service",
        metadata=client.V1ObjectMeta(
            name="{0}-service".format(service_name),
            namespace=service_level,
            labels={"drucker-worker": "True", "app": app_name, "sel": service_name}
        ),
        spec=client.V1ServiceSpec(
            ports=[
                client.V1ServicePort(
                    name="http2",
                    port=service_port,
                    protocol="TCP",
                    target_port=service_port
                )
            ],
            selector={"sel": service_name},
            type="NodePort"
        )
    )
    core_vi = client.CoreV1Api()
    if mode_create:
        api_response = core_vi.create_namespaced_service(
            namespace=service_level,
            body=v1_service
        )
    else:
        api_response = core_vi.patch_namespaced_service(
            namespace=service_level,
            name="{0}-service".format(service_name),
            body=v1_service
        )
    """Ingress"""
    v1_beta1_ingress = client.V1beta1Ingress(
        api_version="extensions/v1beta1",
        kind="Ingress",
        metadata=client.V1ObjectMeta(
            annotations={
                "ingress.kubernetes.io/ssl-passthrough": "true",
                "ingress.zlab.co.jp/backend-config": "{{\"{0}-service\": {{\"{1}\": {{\"proto\":\"h2\"}}}}}}".format(service_name,service_port),
                "kubernetes.io/ingress.class": "nghttpx"
            },
            name="{0}-ingress".format(service_name),
            namespace=service_level,
            labels={"drucker-worker": "True", "app": app_name, "sel": service_name}
        ),
        spec=client.V1beta1IngressSpec(
            rules=[
                client.V1beta1IngressRule(
                    host="{0}-{1}-{2}.{3}".format(app_name, api.dashboard_config.DRUCKER_GRPC_VERSION, service_level, app_dns_name),
                    http=client.V1beta1HTTPIngressRuleValue(
                        paths=[
                            client.V1beta1HTTPIngressPath(
                                backend=client.V1beta1IngressBackend(
                                    service_name="{0}-service".format(service_name),
                                    service_port=service_port
                                )
                            )
                        ]
                    )
                )
            ]
        )
    )
    extensions_v1_beta = client.ExtensionsV1beta1Api()
    if mode_create:
        api_response = extensions_v1_beta.create_namespaced_ingress(
            namespace=service_level,
            body=v1_beta1_ingress
        )
    else:
        api_response = extensions_v1_beta.patch_namespaced_ingress(
            namespace=service_level,
            name="{0}-ingress".format(service_name),
            body=v1_beta1_ingress
        )
    """Autoscaler"""
    v1_horizontal_pod_autoscaler = client.V1HorizontalPodAutoscaler(
        api_version="autoscaling/v1",
        kind="HorizontalPodAutoscaler",
        metadata=client.V1ObjectMeta(
            name="{0}-autoscaling".format(service_name),
            namespace=service_level,
            labels={"drucker-worker": "True", "app": app_name, "sel": service_name}
        ),
        spec=client.V1HorizontalPodAutoscalerSpec(
            max_replicas=replicas_maximum,
            min_replicas=replicas_minimum,
            scale_target_ref=client.V1CrossVersionObjectReference(
                api_version="apps/v1",
                kind="Deployment",
                name="{0}-deployment".format(service_name)
            ),
            target_cpu_utilization_percentage=autoscale_cpu_threshold
        )
    )
    autoscaling_v1 = client.AutoscalingV1Api()
    if mode_create:
        api_response = autoscaling_v1.create_namespaced_horizontal_pod_autoscaler(
            namespace=service_level,
            body=v1_horizontal_pod_autoscaler
        )
    else:
        api_response = autoscaling_v1.patch_namespaced_horizontal_pod_autoscaler(
            namespace=service_level,
            name="{0}-autoscaling".format(service_name),
            body=v1_horizontal_pod_autoscaler
        )
    #TODO: SDK may have a bug? since conditions and current_metrics always are returned NULL, SDK always throws an exception.
    """
    v2_beta1_horizontal_pod_autoscaler = client.V2beta1HorizontalPodAutoscaler(
        api_version="autoscaling/v2beta1",
        kind="HorizontalPodAutoscaler",
        metadata=client.V1ObjectMeta(
            name="{0}-autoscaling".format(service_name),
            namespace=service_level,
            labels={"drucker-worker": "True", "app": app_name, "sel": service_name}
        ),
        spec=client.V2beta1HorizontalPodAutoscalerSpec(
            max_replicas=replicas_maximum,
            min_replicas=replicas_minimum,
            metrics=[
                client.V2beta1MetricSpec(
                    resource=client.V2beta1ResourceMetricSource(
                        name="cpu",
                        target_average_utilization=autoscale_cpu_threshold
                    ),
                    type="Resource"
                )
            ],
            scale_target_ref=client.V2beta1CrossVersionObjectReference(
                api_version="apps/v1",
                kind="Deployment",
                name="{0}-deployment".format(service_name)
            )
        )
    )
    if mode_create:
        api_response = autoscaling_v2_beta1.create_namespaced_horizontal_pod_autoscaler(
            namespace=service_level,
            body=v2_beta1_horizontal_pod_autoscaler
        )
    else:
        api_response = autoscaling_v2_beta1.patch_namespaced_horizontal_pod_autoscaler(
            namespace=service_level,
            name="{0}-autoscaling".format(service_name),
            body=v2_beta1_horizontal_pod_autoscaler
        )
    """
    return service_name

def switch_drucker_service_model_assignment(
        application_id:int, service_id:int, model_id:int):
    """
    Switch model assignment.
    :param application_id:
    :param service_id:
    :param model_id:
    :return:
    """
    mobj = db.session.query(Model).filter(
        Model.application_id==application_id,Model.model_id==model_id).one()
    model_path = mobj.model_path
    sobj = db.session.query(Service).filter(
        Service.application_id == application_id, Service.service_id==service_id).one()
    host = sobj.host
    drucker_dashboard_application = DruckerDashboardClient(logger=api.logger, host=host)
    response_body = drucker_dashboard_application. \
        run_switch_service_model_assignment(model_path=model_path)
    if not response_body.get("status", True):
        raise Exception(response_body.get("message", "Error."))
    sobj.model_id = model_id
    db.session.flush()

    aobj = db.session.query(Application).filter(Application.application_id == application_id).one()
    kubernetes_id = aobj.kubernetes_id
    if kubernetes_id is not None:
        kobj = db.session.query(Kubernetes).filter(Kubernetes.kubernetes_id == kubernetes_id).one()
        config_path = kobj.config_path
        from kubernetes import client, config
        config.load_kube_config(config_path)

        apps_v1 = client.AppsV1Api()
        v1_deployment = apps_v1.read_namespaced_deployment(
            name="{0}-deployment".format(sobj.service_name),
            namespace=sobj.service_level
        )
        for env_ent in v1_deployment.spec.template.spec.containers[0].env:
            if env_ent.name == "DRUCKER_SERVICE_UPDATE_FLAG":
                env_ent.value = "Model switched to model_id={0} at {1:%Y%m%d%H%M%S}".format(model_id, datetime.utcnow())
                break
        api_response = apps_v1.patch_namespaced_deployment(
            body=v1_deployment,
            name="{0}-deployment".format(sobj.service_name),
            namespace=sobj.service_level
        )
    response_body["status"] = True
    return response_body


def dump_drucker_on_kubernetes(
        kobj:Kubernetes, aobj:Application, sobj:Service):
    if kobj is None:
        raise Exception("No such kubernetes_id.")
    if aobj is None:
        raise Exception("No such application_id.")
    if sobj is None:
        raise Exception("No such service_id.")

    config_path = kobj.config_path
    from kubernetes import client, config
    config.load_kube_config(config_path)
    save_dir = pathlib.Path(api.dashboard_config.DIR_KUBE_CONFIG, kobj.display_name, aobj.application_name)
    save_dir.mkdir(parents=True, exist_ok=True)
    api_client = client.ApiClient()

    apps_v1 = client.AppsV1Api()
    v1_deployment = apps_v1.read_namespaced_deployment(
        name="{0}-deployment".format(sobj.service_name),
        namespace=sobj.service_level,
        exact=True,
        export=True
    )
    json.dump(api_client.sanitize_for_serialization(v1_deployment),
              pathlib.Path(
                  api.dashboard_config.DIR_KUBE_CONFIG,
                  kobj.display_name,
                  aobj.application_name,
                  "{0}-deployment.json".format(sobj.service_name)).open("w", encoding='utf-8'),
              ensure_ascii=False, indent=2)
    core_vi = client.CoreV1Api()
    v1_service = core_vi.read_namespaced_service(
        name="{0}-service".format(sobj.service_name),
        namespace=sobj.service_level,
        exact=True,
        export=True
    )
    json.dump(api_client.sanitize_for_serialization(v1_service),
              pathlib.Path(
                  api.dashboard_config.DIR_KUBE_CONFIG,
                  kobj.display_name,
                  aobj.application_name,
                  "{0}-service.json".format(sobj.service_name)).open("w", encoding='utf-8'),
              ensure_ascii=False, indent=2)
    extensions_v1_beta = client.ExtensionsV1beta1Api()
    v1_beta1_ingress = extensions_v1_beta.read_namespaced_ingress(
        name="{0}-ingress".format(sobj.service_name),
        namespace=sobj.service_level,
        exact=True,
        export=True
    )
    json.dump(api_client.sanitize_for_serialization(v1_beta1_ingress),
              pathlib.Path(
                  api.dashboard_config.DIR_KUBE_CONFIG,
                  kobj.display_name,
                  aobj.application_name,
                  "{0}-ingress.json".format(sobj.service_name)).open("w", encoding='utf-8'),
              ensure_ascii=False, indent=2)
    autoscaling_v1 = client.AutoscalingV1Api()
    v1_horizontal_pod_autoscaler = autoscaling_v1.read_namespaced_horizontal_pod_autoscaler(
        name="{0}-autoscaling".format(sobj.service_name),
        namespace=sobj.service_level,
        exact=True,
        export=True
    )
    json.dump(api_client.sanitize_for_serialization(v1_horizontal_pod_autoscaler),
              pathlib.Path(
                  api.dashboard_config.DIR_KUBE_CONFIG,
                  kobj.display_name,
                  aobj.application_name,
                  "{0}-autoscaling.json".format(sobj.service_name)).open("w", encoding='utf-8'),
              ensure_ascii=False, indent=2)
    """
    autoscaling_v2_beta1 = client.AutoscalingV2beta1Api()
    v2_beta1_horizontal_pod_autoscaler = autoscaling_v2_beta1.read_namespaced_horizontal_pod_autoscaler(
        name="{0}-autoscaling".format(sobj.service_name),
        namespace=sobj.service_level,
        exact=True,
        export=True
    )
    json.dump(api_client.sanitize_for_serialization(v2_beta1_horizontal_pod_autoscaler),
              pathlib.Path(
                  api.dashboard_config.DIR_KUBE_CONFIG,
                  kobj.display_name, 
                  aobj.application_name,
                  "{0}-autoscaling.json".format(sobj.service_name)).open("w", encoding='utf-8'),
              ensure_ascii=False, indent=2)
    """


@kube_info_namespace.route('/')
class ApiKubernetes(Resource):
    @kube_info_namespace.marshal_list_with(kube_info)
    def get(self):
        """get_kubernetes"""
        return Kubernetes.query.all()

    @kube_info_namespace.marshal_with(success_or_not)
    def put(self):
        """update_dbs_of_all_kubernetes_app"""
        for kobj in Kubernetes.query.all():
            update_dbs_kubernetes(kobj.kubernetes_id)
        db.session.commit()
        db.session.close()
        response_body = {"status": True, "message": "Success."}
        return response_body

    @kube_info_namespace.marshal_with(success_or_not)
    @kube_info_namespace.expect(kube_file_parser)
    def post(self):
        """add_kubernetes"""
        args = kube_file_parser.parse_args()
        config_path = "{0}/{1}.config".format(api.dashboard_config.DIR_KUBE_CONFIG, uuid.uuid4().hex)
        dns_name = args['dns_name']
        display_name = args['display_name']
        if display_name is None:
            display_name = uuid.uuid4().hex
        host_model_dir = args['host_model_dir']
        pod_model_dir = args['pod_model_dir']
        db_mysql_host = args['db_mysql_host']
        db_mysql_port = args['db_mysql_port']
        db_mysql_dbname = args['db_mysql_dbname']
        db_mysql_user = args['db_mysql_user']
        db_mysql_password = args['db_mysql_password']
        description = args['description']
        newkube = Kubernetes(description=description,
                             config_path=config_path,
                             dns_name=dns_name,
                             display_name=display_name,
                             host_model_dir=host_model_dir,
                             pod_model_dir=pod_model_dir,
                             db_mysql_host=db_mysql_host,
                             db_mysql_port=db_mysql_port,
                             db_mysql_dbname=db_mysql_dbname,
                             db_mysql_user=db_mysql_user,
                             db_mysql_password=db_mysql_password)
        db.session.add(newkube)
        db.session.flush()
        file = args['file']
        file.save(newkube.config_path)

        try:
            from kubernetes import client, config
            config.load_kube_config(newkube.config_path)
            v1 = client.ExtensionsV1beta1Api()
            v1.list_ingress_for_all_namespaces(watch=False)
            update_dbs_kubernetes(newkube.kubernetes_id)
            db.session.commit()
            db.session.close()
            response_body = {"status": True, "message": "Success."}
        except Exception as error:
            os.remove(newkube.config_path)
            raise error
        return response_body


@kube_info_namespace.route('/dump')
class ApiKubernetesDump(Resource):
    @kube_info_namespace.marshal_with(success_or_not)
    def post(self):
        """dump_kubernetes"""
        for kobj in Kubernetes.query.all():
            for aobj in Application.query.filter_by(kubernetes_id=kobj.kubernetes_id).all():
                for sobj in Service.query.filter_by(application_id=aobj.application_id).all():
                    dump_drucker_on_kubernetes(kobj, aobj, sobj)
        response_body = {"status": True, "message": "Success."}
        return response_body


@kube_info_namespace.route('/<int:kubernetes_id>')
class ApiKubernetesId(Resource):
    @kube_info_namespace.marshal_with(kube_info)
    def get(self, kubernetes_id:int):
        """get_kubernetes_id"""
        return Kubernetes.query.filter_by(
            kubernetes_id=kubernetes_id).first_or_404()

    @kube_info_namespace.marshal_with(success_or_not)
    def put(self, kubernetes_id:int):
        """update_dbs_of_kubernetes_app"""
        update_dbs_kubernetes(kubernetes_id)
        db.session.commit()
        db.session.close()
        response_body = {"status": True, "message": "Success."}
        return response_body

    @kube_info_namespace.marshal_with(success_or_not)
    @kube_info_namespace.expect(kube_file_parser)
    def patch(self, kubernetes_id:int):
        """update_kubernetes_id"""
        args = kube_file_parser.parse_args()
        display_name = args['display_name']
        host_model_dir = args['host_model_dir']
        pod_model_dir = args['pod_model_dir']
        db_mysql_host = args['db_mysql_host']
        db_mysql_port = args['db_mysql_port']
        db_mysql_dbname = args['db_mysql_dbname']
        db_mysql_user = args['db_mysql_user']
        db_mysql_password = args['db_mysql_password']
        description = args['description']
        file = args['file']
        dns_name = args['dns_name']

        kobj = db.session.query(Kubernetes).filter(
            Kubernetes.kubernetes_id==kubernetes_id).one()
        if display_name is not None:
            kobj.display_name = display_name
        if description is not None:
            kobj.description = description
        kobj.dns_name = dns_name
        kobj.host_model_dir = host_model_dir
        kobj.pod_model_dir = pod_model_dir
        kobj.db_mysql_host = db_mysql_host
        kobj.db_mysql_port = db_mysql_port
        kobj.db_mysql_dbname = db_mysql_dbname
        kobj.db_mysql_user = db_mysql_user
        kobj.db_mysql_password = db_mysql_password

        config_path = "{0}/{1}".format(api.dashboard_config.DIR_KUBE_CONFIG, uuid.uuid4().hex)
        file.save(config_path)
        try:
            from kubernetes import client, config
            config.load_kube_config(config_path)
            v1 = client.ExtensionsV1beta1Api()
            v1.list_ingress_for_all_namespaces(watch=False)
            os.remove(kobj.config_path)
            shutil.move(config_path, kobj.config_path)
            response_body = {"status": True, "message": "Success."}
            db.session.commit()
            db.session.close()
        except Exception as error:
            os.remove(config_path)
            raise error
        return response_body

    @kube_info_namespace.marshal_with(success_or_not)
    def delete(self, kubernetes_id:int):
        """delete_kubernetes_id"""
        aobj = db.session.query(Application).filter(Application.kubernetes_id == kubernetes_id).all()
        for res in aobj:
            application_id = res.application_id
            db.session.query(Model).filter(Model.application_id == application_id).delete()
            db.session.query(Service).filter(Service.application_id == application_id).delete()
            db.session.query(Application).filter(Application.application_id == application_id).delete()
        db.session.query(Kubernetes).filter(Kubernetes.kubernetes_id == kubernetes_id).delete()
        response_body = {"status": True, "message": "Success."}
        db.session.commit()
        db.session.close()
        return response_body

@kube_info_namespace.route('/<int:kubernetes_id>/applications')
class ApiKubernetesIdApplication(Resource):
    kube_app_deploy = kube_deploy_parser.copy()
    kube_app_deploy.remove_argument('service_model_assignment')

    from .api_application import app_info
    @kube_info_namespace.marshal_list_with(app_info)
    def get(self, kubernetes_id:int):
        """get Kubernetes applications"""
        return Application.query.filter_by(
            kubernetes_id=kubernetes_id).all()

    @kube_info_namespace.expect(kube_app_deploy)
    @kube_info_namespace.marshal_with(success_or_not)
    def post(self, kubernetes_id:int):
        """add_kubernetes_application"""
        args = self.kube_app_deploy.parse_args()
        args["service_model_assignment"] = None
        service_name = create_or_update_drucker_on_kubernetes(kubernetes_id, args)
        update_dbs_kubernetes(kubernetes_id, description=args['commit_message'])
        db.session.commit()
        db.session.close()
        response_body = {"status": True, "message": "Success."}
        return response_body

@kube_info_namespace.route('/<int:kubernetes_id>/applications/<int:application_id>/services')
class ApiKubernetesIdApplicationIdServices(Resource):
    kube_srv_deploy = kube_deploy_parser.copy()
    kube_srv_deploy.remove_argument('app_name')

    from .api_service import srv_info
    @kube_info_namespace.marshal_list_with(srv_info)
    def get(self, kubernetes_id:int, application_id:int):
        """get_kubernetes_service"""
        if db.session.query(Application).filter(
                Application.kubernetes_id == kubernetes_id,
                Application.application_id == application_id).one_or_none() is None:
            raise Exception("No such data.")
        return Service.query.filter_by(application_id=application_id).all()

    @kube_info_namespace.marshal_with(success_or_not)
    def put(self, kubernetes_id:int, application_id:int):
        """update_dbs_of_kubernetes_app"""
        aobj = db.session.query(Application).filter(
            Application.application_id == application_id,
            Application.kubernetes_id == kubernetes_id).one_or_none()
        applist = set((aobj.application_name,))
        update_dbs_kubernetes(kubernetes_id, applist)
        db.session.commit()
        db.session.close()
        response_body = {"status": True, "message": "Success."}
        return response_body

    @kube_info_namespace.expect(kube_srv_deploy)
    @kube_info_namespace.marshal_with(success_or_not)
    def post(self, kubernetes_id:int, application_id:int):
        """add_kubernetes_service"""
        args = self.kube_srv_deploy.parse_args()
        aobj = db.session.query(Application).filter(
            Application.kubernetes_id == kubernetes_id,
            Application.application_id == application_id).one_or_none()
        args["app_name"] = aobj.application_name
        if args["service_model_assignment"] is not None:
            mobj = Model.query.filter_by(
                application_id=application_id,
                model_id=args["service_model_assignment"]).one_or_none()
            if mobj is None:
                args["service_model_assignment"] = None
        service_name = create_or_update_drucker_on_kubernetes(kubernetes_id, args)
        applist = set((aobj.application_name,))
        update_dbs_kubernetes(kubernetes_id, applist, description=args['commit_message'])

        sobj = db.session.query(Service).filter(
            Service.application_id == aobj.application_id,
            Service.service_name == service_name).one_or_none()
        if sobj is not None and sobj.model_id != args["service_model_assignment"]:
            sobj.model_id = args["service_model_assignment"]
            db.session.flush()
        db.session.commit()
        db.session.close()
        response_body = {"status": True, "message": "Success."}
        return response_body

@kube_info_namespace.route('/<int:kubernetes_id>/applications/<int:application_id>/services/<int:service_id>')
class ApiKubernetesIdApplicationIdServiceId(Resource):
    kube_srv_update = kube_deploy_parser.copy()
    kube_srv_update.remove_argument('app_name')
    kube_srv_update.remove_argument('service_level')
    kube_srv_update.remove_argument('service_model_assignment')

    @kube_info_namespace.marshal_with(kube_service_config_info)
    def get(self, kubernetes_id:int, application_id:int, service_id:int):
        """get kubernetes service info"""
        sobj = Service.query.filter_by(
            application_id=application_id,
            service_id=service_id).one_or_none()
        if sobj is None:
            raise Exception("No such data.")
        kobj = Kubernetes.query.filter_by(kubernetes_id=kubernetes_id).one_or_none()
        if kobj is None:
            raise Exception("No such kubernetes_id.")
        config_path = kobj.config_path
        from kubernetes import client, config
        config.load_kube_config(config_path)

        apps_v1 = client.AppsV1Api()
        v1_deployment = apps_v1.read_namespaced_deployment(
            name="{0}-deployment".format(sobj.service_name),
            namespace=sobj.service_level
        )
        autoscaling_v1 = client.AutoscalingV1Api()
        v1_horizontal_pod_autoscaler = autoscaling_v1.read_namespaced_horizontal_pod_autoscaler(
            name="{0}-autoscaling".format(sobj.service_name),
            namespace=sobj.service_level
        )
        """
        autoscaling_v2_beta1 = client.AutoscalingV2beta1Api()
        v2_beta1_horizontal_pod_autoscaler = autoscaling_v2_beta1.read_namespaced_horizontal_pod_autoscaler(
            name="{0}-autoscaling".format(sobj.service_name),
            namespace=sobj.service_level
        )
        """

        response_body = {}
        for env_ent in v1_deployment.spec.template.spec.containers[0].env:
            if env_ent.name == "DRUCKER_SERVICE_UPDATE_FLAG":
                response_body["commit_message"] = env_ent.value
            elif env_ent.name == "DRUCKER_APPLICATION_NAME":
                response_body["app_name"] = env_ent.value
            elif env_ent.name == "DRUCKER_SERVICE_LEVEL":
                response_body["service_level"] = env_ent.value
            elif env_ent.name == "DRUCKER_SERVICE_NAME":
                response_body["service_name"] = env_ent.value
            elif env_ent.name == "DRUCKER_SERVICE_PORT":
                response_body["service_port"] = env_ent.value
            elif env_ent.name == "DRUCKER_SERVICE_GIT_URL":
                response_body["service_git_url"] = env_ent.value
            elif env_ent.name == "DRUCKER_SERVICE_GIT_BRANCH":
                response_body["service_git_branch"] = env_ent.value
            elif env_ent.name == "DRUCKER_SERVICE_BOOT_SHELL":
                response_body["service_boot_script"] = env_ent.value
            elif env_ent.name == "DRUCKER_DB_MYSQL_HOST":
                response_body["db_mysql_host"] = env_ent.value
            elif env_ent.name == "DRUCKER_DB_MYSQL_PORT":
                response_body["db_mysql_port"] = env_ent.value
            elif env_ent.name == "DRUCKER_DB_MYSQL_DBNAME":
                response_body["db_mysql_dbname"] = env_ent.value
            elif env_ent.name == "DRUCKER_DB_MYSQL_USER":
                response_body["db_mysql_user"] = env_ent.value
            elif env_ent.name == "DRUCKER_DB_MYSQL_PASSWORD":
                response_body["db_mysql_password"] = env_ent.value
        response_body["replicas_default"] = v1_deployment.spec.replicas
        response_body["replicas_minimum"] = v1_horizontal_pod_autoscaler.spec.min_replicas
        response_body["replicas_maximum"] = v1_horizontal_pod_autoscaler.spec.max_replicas
        response_body["autoscale_cpu_threshold"] = v1_horizontal_pod_autoscaler.spec.target_cpu_utilization_percentage
        response_body["policy_max_surge"] = v1_deployment.spec.strategy.rolling_update.max_surge
        response_body["policy_max_unavailable"] = v1_deployment.spec.strategy.rolling_update.max_unavailable
        response_body["policy_wait_seconds"] = v1_deployment.spec.min_ready_seconds
        response_body["container_image"] = v1_deployment.spec.template.spec.containers[0].image
        response_body["resource_request_cpu"] = kubernetes_cpu_to_float(v1_deployment.spec.template.spec.containers[0].resources.requests["cpu"])
        response_body["resource_request_memory"] = v1_deployment.spec.template.spec.containers[0].resources.requests["memory"]
        response_body["resource_limit_cpu"] = kubernetes_cpu_to_float(v1_deployment.spec.template.spec.containers[0].resources.limits["cpu"])
        response_body["resource_limit_memory"] = v1_deployment.spec.template.spec.containers[0].resources.limits["memory"]
        response_body["host_model_dir"] = v1_deployment.spec.template.spec.volumes[0].host_path.path
        response_body["pod_model_dir"] = v1_deployment.spec.template.spec.containers[0].volume_mounts[0].mount_path
        return response_body


    @kube_info_namespace.expect(kube_srv_update)
    @kube_info_namespace.marshal_with(success_or_not)
    def patch(self, kubernetes_id:int, application_id:int, service_id:int):
        """update_kubernetes_deployment"""
        args = self.kube_srv_update.parse_args()
        sobj = db.session.query(Service).filter(
            Service.service_id == service_id).one_or_none()
        if sobj is None:
            raise Exception("No such data.")
        aobj = db.session.query(Application).filter(
            Application.kubernetes_id == kubernetes_id,
            Application.application_id == application_id).one_or_none()
        if aobj is None:
            raise Exception("No such data.")
        args["app_name"] = aobj.application_name
        args["service_level"] = sobj.service_level
        args['commit_message'] = "Request a rolling-update at {0:%Y%m%d%H%M%S}".format(datetime.utcnow())
        args["service_model_assignment"] = None
        create_or_update_drucker_on_kubernetes(kubernetes_id, args, sobj.service_name)
        applist = set((aobj.application_name,))
        update_dbs_kubernetes(kubernetes_id, applist)
        db.session.commit()
        db.session.close()
        response_body = {"status": True, "message": "Success."}
        return response_body



# coding: utf-8


from .dao import db
from .kubernetes import Kubernetes
from .application import Application
from .service import Service
from .model import Model
from .user import User
from .application_user_role import ApplicationUserRole, Role
from .evaluation import Evaluation
from .evaluation_result import EvaluationResult



import datetime
from .dao import db
from sqlalchemy import (
    Column, Integer, String, DateTime,
    Text, UniqueConstraint
)


class Service(db.Model):
    """
    Services
    """
    __tablename__ = 'services'
    __table_args__ = (
        UniqueConstraint('service_id'),
        UniqueConstraint('service_name'),
        UniqueConstraint('application_id','display_name'),
        {'mysql_engine': 'InnoDB'}
    )

    service_id = Column(Integer, primary_key=True, autoincrement=True)
    service_name = Column(String(512), nullable=False)
    display_name = Column(String(128), nullable=False)
    application_id = Column(Integer, nullable=False)
    model_id = Column(Integer, nullable=True)
    service_level = Column(String(128), nullable=False)
    register_date = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)
    confirm_date = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)
    update_date = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)
    host = Column(String(512), nullable=False)
    description = Column(Text, nullable=True)

    @property
    def serialize(self):
        return {
            'service_id': self.service_id,
            'service_name': self.service_name,
            'display_name': self.display_name,
            'application_id': self.application_id,
            'model_id': self.model_id,
            'service_level': self.service_level,
            'register_date': self.register_date.strftime('%Y-%m-%d %H:%M:%S'),
            'confirm_date': self.confirm_date.strftime('%Y-%m-%d %H:%M:%S'),
            'update_date': self.update_date.strftime('%Y-%m-%d %H:%M:%S'),
            'host': self.host,
            'description': self.description,
        }
